--[[

TidyShop - Cleans up the in-game store for a tidier and more uniform look

Author:     w33zl (https://github.com/w33zl)
Version:    1.0
Modified:   2020-10-16

Changelog:
v1.0    Initial release

Copyright (C) 2020 by w33zl

This mod is ONLY intended to be distributed via the Giants ModHub (https://farming-simulator.com/mods.php). 
It is strictly forbidden to upload this software/mod to any other third-party download service that offers money in exchange of downloads!

If this mod is made available trough any other website or distribution media (other than the Giant's ModHub) 
this website/media must reference back to the original location at Giant's ModHub and clearly state the origin of the mod.

]]

TidyShop = Mod:init()

TidyShop.debugMode = false

TidyShop:source("categories.lua");
TidyShop:source("storeitems.lua");

local DEFAULT_IMAGE = "images/cat_default.dds"

local MatchableItem = {}
MatchableItem.__index = MatchableItem

function TidyShop:loadMap(savegame)
    -- Init the mod
    self:init()

    -- If GlobalCompany is present, delay execution of load event (otherwise execute immediately)
    if g_company ~= nil and type(g_company.addLoadable) == "function" then
        g_company.addLoadable(self, self.load)
    else
        self:load()
    end
end

function TidyShop:init()

    -- Init mod settings
    local settings = self.settings:init("tidyShop", "settings.xml", "tidyShop.xml");

    -- Set initial default values
    settings.leaveCustomDlcTitles = false;
    settings.allowReplacingCategories = true;
    settings.rearrangeCategories = true;
    settings.categories = {};

    -- Load actual settings from both the default settings file from the mod and user settings from the modsSettings-folder
    settings:load(function(xmlReader)
        settings.leaveCustomDlcTitles = xmlReader:readBool(nil, "leaveCustomDlcTitles", settings.leaveCustomDlcTitles)
        settings.allowReplacingCategories = xmlReader:readBool(nil, "allowReplacingCategories", settings.allowReplacingCategories)

        settings.categories = {
            mergeSwapBodyVehicles = xmlReader:readBool("categories", "mergeSwapBodyVehicles", settings.categories.mergeSwapBodyVehicles),
            mergeHookliftVehicles = xmlReader:readBool("categories", "mergeHookliftVehicles", settings.categories.mergeHookliftVehicles),
            moveLiquidsSemitrailersToGenericSemitrailers = xmlReader:readBool("categories", "moveLiquidsSemitrailersToGenericSemitrailers", settings.categories.moveLiquidsSemitrailersToGenericSemitrailers),
            moveBulkSemitrailersToGenericSemitrailers = xmlReader:readBool("categories", "moveBulkSemitrailersToGenericSemitrailers", settings.categories.moveBulkSemitrailersToGenericSemitrailers),
            moveFlatbedSemitrailersToGenericSemitrailers = xmlReader:readBool("categories", "moveFlatbedSemitrailersToGenericSemitrailers", settings.categories.moveFlatbedSemitrailersToGenericSemitrailers),
        }
    end);

    -- -- This will auto-save all default settings to the user settings file (in modsSettings-folder) on the first run
    -- self.settings:save(function(xmlWriter)
    --     xmlWriter:saveBool(nil, "leaveCustomDlcTitles", settings.leaveCustomDlcTitles)
    --     xmlWriter:saveBool("", "allowReplacingCategories", settings.allowReplacingCategories)

    --     xmlWriter:saveBool("categories", "mergeSwapBodyVehicles", settings.categories.mergeSwapBodyVehicles)
    --     xmlWriter:saveBool("categories", "mergeHookliftVehicles", settings.categories.mergeHookliftVehicles)
    --     xmlWriter:saveBool("categories", "moveLiquidsSemitrailersToGenericSemitrailers", settings.categories.moveLiquidsSemitrailersToGenericSemitrailers)
    --     xmlWriter:saveBool("categories", "moveBulkSemitrailersToGenericSemitrailers", settings.categories.moveBulkSemitrailersToGenericSemitrailers)
    --     xmlWriter:saveBool("categories", "moveFlatbedSemitrailersToGenericSemitrailers", settings.categories.moveFlatbedSemitrailersToGenericSemitrailers)
    -- end);
end

-- Load Event (after init)
function TidyShop:load()
    self:handleCategories()
    self:handleStoreItems()
    self:postfix()
end

-- Eventueal fixes to be executed at the very end
function TidyShop:postfix()
    --TODO: clean empty categories
end

function TidyShop:handleStoreItems()
    if self.alterStoreItem == nil or type(self.alterStoreItem) ~= "function" then
        self:printWarning("No TidyShop:alterStoreItem() was found, skipping changing storeitems!")
        return
    elseif g_storeManager == nil then
        self:printError("Store manager (g_storeManager) is unavable, aborting!");
        return
    end

    for _, item in pairs(g_storeManager:getItems()) do
        self:alterStoreItem(MatchableItem:new(item));
    end
end

function TidyShop:handleCategories()
    if g_storeManager == nil then
        self:printError("Store manager (g_storeManager) is unavable, aborting!");
        return
    end

    self:rearrangeCategories();

    if self.initializeCategories ~= nil and type(self.initializeCategories) == "function" then
        -- self:printDebugVar("Categories before", g_currentMission.shopController.storeManager.numOfCategories)

        self:initializeCategories()

        -- self:printDebugVar("Categories after", g_currentMission.shopController.storeManager.numOfCategories)
    else
        self:printWarning("No TidyShop:initializeCategories() was found, skipping adding categories!")
    end
end

function TidyShop:rearrangeCategories()
    if self.settings.rearrangeCategories ~= true then
        return;
    end

    --TODO: re arrange (ok to add +1000, maybe sort per main category? 1000... 2000... 3000... 4000...)

    --HACK: quick-fix to get enough gap between each category for convinient reordering
    for key, category in pairs(g_storeManager.categories) do
        category.orderId = (category.orderId * 5);
    end
end

function TidyShop:validateParam(value, typeName, message)
    local failed = false
    failed = failed or (value == nil)
    failed = failed or (typeName ~= nil and type(value) ~= typeName)
    failed = failed or (type(value) == string and value == "")

    if failed and message ~= nil and message ~= "" then self:printError(message) end

    return not failed
end

local function getL10nKeyFromDefaultTitle(categoryTitle)
    local autoName = categoryTitle
    local x, y = "%(", "%)"
    
    autoName = autoName:gsub("&", "And")
    autoName = autoName:gsub("/", "And")
    
    autoName = autoName:gsub(",", "")
    autoName = autoName:gsub("-", "")
    autoName = autoName:gsub("(buy)", "")
    autoName = autoName:gsub("(sell)", "")
    autoName = autoName:gsub("(income)", "")
    autoName = autoName:gsub(x, "")
    autoName = autoName:gsub(y, "")
    autoName = autoName:gsub(" ", "")

    return "cat_" .. autoName
end--function

local function getTranslatedTitleWithFallback(categoryTitle)
    local translationKey = getL10nKeyFromDefaultTitle(categoryTitle)
    local transTitle = g_i18n:getText(translationKey)

    if transTitle == nil or transTitle:sub(1, 9) == "Missing '" then
        TidyShop:printWarning("Could not find any l10n text with the name '" .. translationKey .. "', using fallback '" .. categoryTitle .. "'")
        return categoryTitle -- Fallback
    else
        return transTitle
    end
end--function

function TidyShop:addCategory(categoryType, newCategoryName, newCategoryTitle, imagePath, siblingCategoryName, categoryIndexOffset)
    -- self:printDebug(string.format( "\nRAW: Add category '%s' with image path '%s' with sibling '%s' and with a position offset of #%s", newCategoryName, imagePath, siblingCategoryName, categoryIndexOffset ))

    if not self:validateParam(categoryType, "string", "Parameter 'categoryType' (#1) is mandatory and must contain a non-empty string.") then return end
    if not self:validateParam(newCategoryName, "string", "Parameter 'newCategoryName' (#2) is mandatory and must contain a non-empty string.") then return end
    if not self:validateParam(newCategoryTitle, "string", "Parameter 'newCategoryTitle' (#3) is mandatory and must contain a non-empty string.") then return end

    categoryType = categoryType:upper();
    if categoryType ~= "VEHICLE" and categoryType ~= "TOOL" and categoryType ~= "OBJECT" and categoryType ~= "PLACEABLE" then
        self:printError("Parameter 'categoryType' (#1) has an invalid value of '"  .. categoryType .. "'. Allowed values are 'VEHICLE','TOOL', 'OBJECT' and 'PLACEABLE'");
        return;
    end

    -- Translate the title
    newCategoryTitle = getTranslatedTitleWithFallback(newCategoryTitle)

    -- Make sure we have proper default/fallback values
    newCategoryName = newCategoryName:upper();
    imagePath = imagePath or DEFAULT_IMAGE;
    siblingCategoryName = siblingCategoryName or "" --"MISC"; -- TODO: maybe set different default per TYTPE?
    categoryIndexOffset = categoryIndexOffset or 0;

    local existingCategory = g_storeManager:getCategoryByName(newCategoryName);
    local siblingCategory = g_storeManager:getCategoryByName(siblingCategoryName) --or g_storeManager:getCategoryByName("MISC") -- Get sibling category or "MISC" as a fallback
    local isReplacing = (existingCategory ~= nil);
    local categoryModDir = self.dir;
    local newCategoryOrderId = 0
    local siblingCategoryOrderId

    -- Get orderId for sibling category
    if siblingCategory ~= nil then
        siblingCategoryOrderId = siblingCategory.orderId
    end

    -- Enable replacement of existing category by removing the old one
    if isReplacing then
        if self.settings.allowReplacingCategories == true then

            siblingCategoryOrderId = siblingCategoryOrderId or existingCategory.orderId -- Save existing category's orderId as a fallback if we couldn't find one for the sibling category

            g_storeManager:removeCategory(newCategoryName);
            local postDeleteCategory = g_storeManager:getCategoryByName(newCategoryName);

            if postDeleteCategory ~= nil then
                self:printWarning(string.format( "Failed to remove category ", newCategoryName ));
            end
        else
            self:printWarning(string.format("Replacing categories is not allowed, skipping category '%s' due to conflict with existing category. Consider changing the setting <allowReplacingCategories /> in the config.", newCategoryName));
            return;
        end
    end


    -- Calculate the new orderId
    siblingCategoryOrderId = (siblingCategoryOrderId or 0)
    newCategoryOrderId = siblingCategoryOrderId + categoryIndexOffset


    -- Calculate the correct image path and modDirectory
    if imagePath:sub(1, 1) == "$" then
        if imagePath:sub(1, 5) == "$data" then
            imagePath = imagePath:sub(2)
            categoryModDir = ""
        else
            self:printWarning(string.format( "Invalid path '%s', using the default image instead.", imagePath))
            imagePath = DEFAULT_IMAGE
        end
    end

    -- Fallback if image file does not exists
    local doesImageFileExists = fileExists(categoryModDir .. imagePath)

    if not doesImageFileExists then
        imagePath = DEFAULT_IMAGE
        categoryModDir = self.dir
    end

    local success = g_storeManager:addCategory(newCategoryName, newCategoryTitle, imagePath, categoryType, categoryModDir);

    if success then
        local newlyCreatedCategory = g_storeManager:getCategoryByName(newCategoryName)
        success = success and (newlyCreatedCategory ~= nil)
        if success then -- If everything seems ok, update the orderId of the new category
            newlyCreatedCategory.orderId = newCategoryOrderId
        end
    end

    if success then
        if isReplacing then
            self:printInfo(string.format( "Successfully replaced category '%s'", newCategoryName))
        else
            self:printInfo(string.format( "Successfully added category '%s'", newCategoryName))
        end
    else
        self:printError(string.format( "Failed to add category '%s'", newCategoryName))
    end
end--function addCategory

-- --- Fake-version override of method addCategory to automatically generate l10n texts for the xml file
-- TidyShop.addCategory = function(self, categoryType, newCategoryName, newCategoryTitle, imagePath, siblingCategoryName, categoryIndexOffset)
--     local fixedTitle = newCategoryTitle:gsub("&", "&amp;")

--     self.fakeAddCategory_translationKeyIndex = self.fakeAddCategory_translationKeyIndex or {} -- Keep track of already added items

--     local translationbKey = getL10nKeyFromDefaultTitle(newCategoryTitle)

--     -- self:printDebugVar("fakeAddCategory_translationKeyIndex", TidyShop.fakeAddCategory_translationKeyIndex[translationbKey])

--     if self.fakeAddCategory_translationKeyIndex[translationbKey] == nil then
--         print("<text name=\"" .. translationbKey .. "\" text=\"" .. fixedTitle .. "\" />")
--         self.fakeAddCategory_translationKeyIndex[translationbKey] = fixedTitle
--     -- else
--     --     self:printDebugVar("Key already added", translationbKey)
--     end--if
-- end--function

function TidyShop:moveCategory(categoryName, previousCategoryName, indexDelta)
	local baseOrderId = 0;

	if previousCategoryName ~= nil and previousCategoryName ~= "" then
		previousCategory = g_storeManager:getCategoryByName(previousCategoryName);

		if previousCategory ~= nil then
			baseOrderId = previousCategory.orderId;
		end--if
	end--if

	category = g_storeManager:getCategoryByName(categoryName);

	if category ~= nil then
		category.orderId = baseOrderId + indexDelta;
	end--if


end--function

local function stringMatch(s1, s2)
	return (string.find(s1, s2, 1, true) ~= nil);
end--function
		

function MatchableItem:new(storeItem)
    local matchedStoreItem = {}
    setmetatable(matchedStoreItem, MatchableItem);
    matchedStoreItem.storeItem = storeItem;
    matchedStoreItem.lastMatchFailed = false;
    return matchedStoreItem
end--function

function MatchableItem:match()
    local matchedStoreItem = {}
    setmetatable(matchedStoreItem, MatchableItem);
    matchedStoreItem.storeItem = self.storeItem;
    matchedStoreItem.lastMatchFailed = false;
    --TODO: add blacklist/exclusions
    return matchedStoreItem
end--function

function MatchableItem:logResult(name, result)

    self.log = (self.log or "");

    if self.log ~= "" then
        self.log = self.log .. " > ";
    end--if

    self.log = self.log .. name .. "=" .. tostring(result);
end--function


function MatchableItem:debugWarning(message)
    if not DEBUG_MODE then return end

    self:debug("Warning: " .. message);
end--function

function MatchableItem:debug(message)
    if not DEBUG_MODE then return end

    Mod:printDebug(message);
end--function


function MatchableItem:hasName(itemName, exactMatch)

    if self.lastMatchFailed == true then
        return self;
    end--if

    local fixedStoreName = string.lower(self.storeItem.name);
    local fixedName = string.lower(itemName);

    exactMatch = (exactMatch or false) and true
    
    local function doMatch()
        if exactMatch == true then
            return fixedStoreName == fixedName
        else 
            return stringMatch(fixedStoreName, fixedName)
        end
    end
    local isMatch = doMatch()

    self.lastMatchFailed = not isMatch;

    self:logResult("hasName '" .. itemName .. "'", isMatch);

    return self;
end--function

--TODO: addSpec

--TODO: duplicate

--TODO: hasBrand

--TODO: changeBrand / newBrand

function MatchableItem:isBaseGame()
    if self.lastMatchFailed == true then return self end--if

    local isMatch = (self.storeItem.customEnvironment == nil or self.storeItem.customEnvironment == "");

    self.lastMatchFailed = not isMatch;

    self:logResult("isBaseGame", isMatch);

    return self;
end--function	

function MatchableItem:isMod(modName)
    if self.lastMatchFailed == true then return self end--if

    local isMatch = false;
    
    if modName ~= nil and modName ~= "" then
        if self.storeItem.customEnvironment ~= nil and self.storeItem.customEnvironment ~= "" then
            local fixedStoreModName = string.lower(self.storeItem.customEnvironment);
            local fixedModName = string.lower(modName);
            isMatch = stringMatch(fixedStoreModName, fixedModName);
        end--if
    else
        self:debugWarning("Parameter 'modName' is required!");
    end--if

    self.lastMatchFailed = not isMatch;

    self:logResult("isMod '" .. modName .. "'", isMatch);

    return self;
end--function

function MatchableItem:inCategory(category)
    if self.lastMatchFailed == true then return self end--if

    local isMatch = false;
    
    if category ~= nil and category ~= "" then
        local fixedStoreCategoryName = string.upper(self.storeItem.categoryName);
        local fixedMatchCategoryName = string.upper(category);
        isMatch = (fixedStoreCategoryName == fixedMatchCategoryName);
    else
        self:debugWarning("Parameter 'category' is required!");
    end--if

    self.lastMatchFailed = not isMatch;

    self:logResult("inCategory '" .. category .. "'", isMatch);

    return self;
end--function

function MatchableItem:hasFilename(filename)
    if self.lastMatchFailed == true then return self end--if

    local isMatch = false;
    
    if filename ~= nil and filename ~= "" then
        local fixedStoreFileName = string.lower(self.storeItem.xmlFilenameLower);
        local fixedMatchFileName = string.lower(filename);
        if fixedMatchFileName:sub(-4) ~= ".xml" then
            fixedMatchFileName = fixedMatchFileName .. ".xml";
        end--if
        isMatch = stringMatch(fixedStoreFileName, fixedMatchFileName);
    else
        self:debugWarning("Parameter 'filename' is required!");
    end--if

    self.lastMatchFailed = not isMatch;

    self:logResult( "hasFilename '" .. filename .. "'", isMatch);

    return self;
end--function

function MatchableItem:hasSpecs(specs)
    if self.lastMatchFailed == true then return self end--if

    local isMatch = false;
    
    if specs ~= nil and specs ~= "" then
        local storeItem = self.storeItem;

        if storeItem ~= nil and storeItem.specs ~= nil and storeItem.specs.combination ~= nil then
            local fixedSpecsList = string.lower(storeItem.specs.combination);
            local fixedMatchSpecs = string.lower(specs);
            isMatch = stringMatch(fixedSpecsList, fixedMatchSpecs);
        end--if

    else
        self:debugWarning("Parameter 'specs' is required!");
    end--if

    self.lastMatchFailed = not isMatch;

    self:logResult("hasSpecs '" .. specs .. "'", isMatch);

    return self;
end--function

function MatchableItem:getSpecType(specTypeName)

    local specTypePassiveIncome = g_storeManager:getSpecTypeByName(specTypeName)
    local success, passiveIncome = pcall(specTypePassiveIncome.getValueFunc, self.storeItem)

    if success then
        return passiveIncome
    else
        return nil
    end
end

function MatchableItem:hasCombo(name)
    if self.lastMatchFailed == true then return self end--if

    local combination = self:getSpecType("combination")
    local isMatch = false
    
    if combination ~= nil then
        local fixedStoreComboList = string.lower(combination);
        local fixedMatchComboList = string.lower(name);
        isMatch = stringMatch(fixedStoreComboList, fixedMatchComboList);
    end

    self.lastMatchFailed = not isMatch;

    self:logResult("hasCombo", isMatch);

    return self;
end

--TODO: add to hasXxxx()
-- 		specs :: table: 0x019ef7b98938
-- 2020-07-08 20:29 .        seedFillTypes :: table: 0x019ef69579d0
-- 2020-07-08 20:29 .        fillTypes :: table: 0x019ef6672af8
-- 2020-07-08 20:29 .        fuel :: table: 0x019ef6672a00
-- 2020-07-08 20:29 .        capacity :: table: 0x019ef6672ab0
-- 2020-07-08 20:29 .        workingWidthVar :: table: 0x019ef73627d8
-- 2020-07-08 20:29 .        incomePerHour :: table: 0x019ef7027ec8
-- 2020-07-08 20:29 .707 :: table: 0x019ef6782120
-- 2020-07-08 20:29 .    specs :: table: 0x019ef77e1560
-- 2020-07-08 20:29 .        seedFillTypes :: table: 0x019ef6782200
-- 2020-07-08 20:29 .        fillTypes :: table: 0x019ef6c4d000
-- 2020-07-08 20:29 .        capacity :: table: 0x019ef6c4d068
-- 2020-07-08 20:29 .        power :: 200
-- 2020-07-08 20:29 .        fuel :: table: 0x019ef73c61d0
-- 2020-07-08 20:29 .        maxSpeed :: 80
-- 2020-07-08 20:29 .        workingWidthVar :: table: 0x019ef6938a18
-- 2020-07-08 21:27 [ATidierStore] Passive income :: Greenhouse : cucumber
-- 2020-07-08 21:27   loadFunc :: function: 0x01c6d7c84200
-- 2020-07-08 21:27   getValueFunc :: function: 0x01c6d7c83928
-- 2020-07-08 21:27   name :: incomePerHour
-- 2020-07-08 21:27   profile :: shopListAttributeIconIncomePerHour
-- 2020-07-08 21:27   seedFillTypes :: table: 0x01c6b86487f8
-- 2020-07-08 21:27   fillTypes :: table: 0x01c6b726ed60
-- 2020-07-08 21:27       useWindrowed :: false
-- 2020-07-08 21:27   fuel :: table: 0x01c6b2ce9510
-- 2020-07-08 21:27       fillUnits :: table: 0x01c6b86488b0
-- 2020-07-08 21:27       consumers :: table: 0x01c6b86488f8
-- 2020-07-08 21:27   capacity :: table: 0x01c6b726ed18
-- 2020-07-08 21:27   workingWidthVar :: table: 0x01c6b726eda8
-- 2020-07-08 21:27   incomePerHour :: table: 0x01c6bcf50598
-- 2020-07-08 21:27       1 :: 88
-- 2020-07-08 21:27       2 :: 56
-- 2020-07-08 21:27       3 :: 14

function MatchableItem:hasPassiveIncome()
    if self.lastMatchFailed == true then return self end--if

    local passiveIncome = self:getSpecType("incomePerHour")
    local isMatch = passiveIncome ~= nil

    self.lastMatchFailed = not isMatch;

    self:logResult("hasPassiveIncome", isMatch);

    return self;
end--function	

function MatchableItem:hasPower(min, max)
    if self.lastMatchFailed == true then return self end--if
end

function MatchableItem:moveTo(newCategory)
    if self.lastMatchFailed == true then return self end--if

    if newCategory ~= nil and newCategory ~= "" and g_storeManager:getCategoryByName(newCategory) ~= nil then
        self.storeItem.categoryName = string.upper(newCategory);
    else
        self.lastMatchFailed = true;
        self:debugWarning("Could not find category '" .. newCategory .. "', aborting move!");
    end--if

    self:logResult("moveTo", not self.lastMatchFailed);

    return self;
end--function

function MatchableItem:newCategory(newCategory)
    self:moveTo(newCategory);
end

function MatchableItem:dump(newCategory)

    local matchString

    if self.lastMatchFailed then 
        matchString = "NOT a"
    else 
        matchString = "a"
    end

    self:debug(string.format( "MatchableItem:dump :: Name='%s' | Category=%s | Filename='%s' was %s match: %s", self.storeItem.name, self.storeItem.categoryName, self.storeItem.xmlFilenameLower, matchString, self.log));

    return self;
end--function

function MatchableItem:rename(newName)
    if self.lastMatchFailed == true then return self end--if

    if newName ~= nil and newName ~= "" then
        self.storeItem.name = newName;
    end;

end--function

function MatchableItem:addSuffix(suffix)
    if self.lastMatchFailed == true then return self end--if

    if suffix ~= nil and suffix ~= "" then
        self.storeItem.name = self.storeItem.name .. suffix;
    end;	
end--function

function MatchableItem:fixDlcTitle(forceChange)
    if self.lastMatchFailed == true then return self end--if

    forceChange = forceChange or true

    local modName = self.storeItem.customEnvironment or ""
    local shouldUpdate = (modName ~= "") and (self.storeItem.isMod or forceChange)

    if shouldUpdate then
        modName = modName:gsub("FS19_", "")
        modName = modName:gsub("LS19_", "")
        modName = modName:gsub("_", " ")
        self.storeItem.dlcTitle = modName
        self.storeItem.isMod = false
    end

    --TODO: change title

end--function
